@@ -50,15 +50,8 @@ test_that("can interpolate in specific env", {
 })
 
 test_that("can qualify operators with namespace", {
-  with_non_verbose_retirement({
-    # Should remove prefix only if rlang-qualified:
-    expect_identical(quo(rlang::UQ(toupper("a"))), new_quosure("A", empty_env()))
-    expect_identical(quo(list(rlang::UQS(list(a = 1, b = 2)))), quo(list(a = 1, b = 2)))
-
-    # Should keep prefix otherwise:
     expect_identical(quo(other::UQ(toupper("a"))), quo(other::"A"))
     expect_identical(quo(x$UQ(toupper("a"))), quo(x$"A"))
-  })
 })
 
 test_that("unquoting is frame-consistent", {
@@ -164,7 +157,7 @@ test_that("quosures are not rewrapped", {
   var <- quo(!! quo(letters))
   expect_identical(quo(!!var), quo(letters))
 
-  var <- new_quosure(local(~letters), env = child_env(get_env()))
+  var <- new_quosure(local(~letters), env = child_env(current_env()))
   expect_identical(quo(!!var), var)
 })
 
@@ -173,27 +166,15 @@ test_that("UQ() fails if called without argument", {
     quo <- quo(UQ(NULL))
     expect_equal(quo, ~NULL)
 
-    quo <- quo(rlang::UQ(NULL))
-    expect_equal(quo, ~NULL)
-
     quo <- tryCatch(quo(UQ()), error = identity)
     expect_is(quo, "error")
     expect_match(quo$message, "must be called with an argument")
-
-    quo <- tryCatch(quo(rlang::UQ()), error = identity)
-    expect_is(quo, "error")
-    expect_match(quo$message, "must be called with an argument")
   })
 })
 
 
 # !!! ---------------------------------------------------------------------
 
-test_that("`!!!` treats atomic objects as scalar vectors", {
-  expect_identical(quo(list(!!! get_env())), quo(list(!! get_env())))
-  expect_identical(expr(c(!!! expression(1, 2))), expr(c(!! expression(1, 2))))
-})
-
 test_that("values of `!!!` spliced into expression", {
   f <- quo(f(a, !!! list(quote(b), quote(c)), d))
   expect_identical(f, quo(f(a, b, c, d)))
@@ -234,10 +215,10 @@ test_that("can't splice at top level", {
 
 test_that("can splice function body even if not a `{` block", {
   fn <- function(x) { x }
-  expect_identical(exprs(!!! body(fn)), named_list(quote(x)))
+  expect_identical(exprs(!!!fn_body(fn)), named_list(quote(x)))
 
   fn <- function(x) x
-  expect_identical(exprs(!!! body(fn)), named_list(quote(x)))
+  expect_identical(exprs(!!!fn_body(fn)), named_list(quote(x)))
 })
 
 test_that("splicing a pairlist has no side effect", {
@@ -281,23 +262,112 @@ test_that("`!!!` doesn't modify spliced inputs by reference", {
   expect_equal(x, quote({ 1L; 2L; 3L }))  # equal because of srcrefs
 })
 
-
-# UQE ----------------------------------------------------------------
-
-test_that("UQE() extracts right-hand side", {
-  var <- ~cyl
-  expect_warning_(expect_identical_(quo(mtcars$UQE(var)), quo(mtcars$cyl)), "deprecated")
+test_that("exprs() preserves spliced quosures", {
+  out <- exprs(!!!quos(a, b))
+  expect_identical(out, exprs(!!quo(a), !!quo(b)))
+  expect_identical(out, named_list(quo(a), quo(b)))
 })
 
-test_that("UQE() throws a deprecation warning", {
-  expect_warning_(exprs(UQE("foo")), "deprecated")
-  expect_warning_(quos(UQE("foo")), "deprecated")
-  expect_warning_(expr(UQE("foo")), "deprecated")
-  expect_warning_(quo(UQE("foo")), "deprecated")
+test_that("!!! fails with non-vectors", {
+  expect_error_(exprs(!!!env()), "not a vector")
+  expect_error_(exprs(!!!function() NULL), "not a vector")
+  expect_error_(exprs(!!!base::c), "not a vector")
+  expect_error_(exprs(!!!base::`{`), "not a vector")
+  expect_error_(exprs(!!!expression()), "not a vector")
+
+  expect_error_(quos(!!!env()), "not a vector")
+  expect_error_(quos(!!!function() NULL), "not a vector")
+  expect_error_(quos(!!!base::c), "not a vector")
+  expect_error_(quos(!!!base::`{`), "not a vector")
+  expect_error_(quos(!!!expression()), "not a vector")
+
+  expect_error_(expr(list(!!!env())), "not a vector")
+  expect_error_(expr(list(!!!function() NULL)), "not a vector")
+  expect_error_(expr(list(!!!base::c)), "not a vector")
+  expect_error_(expr(list(!!!base::`{`)), "not a vector")
+  expect_error_(expr(list(!!!expression())), "not a vector")
+
+  expect_error_(list2(!!!env()), "not a vector")
+  expect_error_(list2(!!!function() NULL), "not a vector")
+  expect_error_(list2(!!!base::c), "not a vector")
+  expect_error_(list2(!!!base::`{`), "not a vector")
+  expect_error_(list2(!!!expression()), "not a vector")
 })
 
-test_that("UQE() can't be used in by-value dots", {
-  expect_error_(dots_list(UQE("foo")), "non-quoting function")
+test_that("!!! succeeds with vectors, pairlists and language objects", {
+  expect_identical_(exprs(!!!NULL), named_list())
+  expect_identical_(exprs(!!!pairlist(1)), named_list(1))
+  expect_identical_(exprs(!!!list(1)), named_list(1))
+  expect_identical_(exprs(!!!TRUE), named_list(TRUE))
+  expect_identical_(exprs(!!!1L), named_list(1L))
+  expect_identical_(exprs(!!!1), named_list(1))
+  expect_identical_(exprs(!!!1i), named_list(1i))
+  expect_identical_(exprs(!!!"foo"), named_list("foo"))
+  expect_identical_(exprs(!!!bytes(0)), named_list(bytes(0)))
+
+  expect_identical_(quos(!!!NULL), quos_list())
+  expect_identical_(quos(!!!pairlist(1)), quos_list(quo(1)))
+  expect_identical_(quos(!!!list(1)), quos_list(quo(1)))
+  expect_identical_(quos(!!!TRUE), quos_list(quo(TRUE)))
+  expect_identical_(quos(!!!1L), quos_list(quo(1L)))
+  expect_identical_(quos(!!!1), quos_list(quo(1)))
+  expect_identical_(quos(!!!1i), quos_list(quo(1i)))
+  expect_identical_(quos(!!!"foo"), quos_list(quo("foo")))
+  expect_identical_(quos(!!!bytes(0)), quos_list(quo(!!bytes(0))))
+
+  expect_identical_(expr(foo(!!!NULL)), quote(foo()))
+  expect_identical_(expr(foo(!!!pairlist(1))), quote(foo(1)))
+  expect_identical_(expr(foo(!!!list(1))), quote(foo(1)))
+  expect_identical_(expr(foo(!!!TRUE)), quote(foo(TRUE)))
+  expect_identical_(expr(foo(!!!1L)), quote(foo(1L)))
+  expect_identical_(expr(foo(!!!1)), quote(foo(1)))
+  expect_identical_(expr(foo(!!!1i)), quote(foo(1i)))
+  expect_identical_(expr(foo(!!!"foo")), quote(foo("foo")))
+  expect_identical_(expr(foo(!!!bytes(0))), expr(foo(!!bytes(0))))
+
+  expect_identical_(list2(!!!NULL), list())
+  expect_identical_(list2(!!!pairlist(1)), list(1))
+  expect_identical_(list2(!!!list(1)), list(1))
+  expect_identical_(list2(!!!TRUE), list(TRUE))
+  expect_identical_(list2(!!!1L), list(1L))
+  expect_identical_(list2(!!!1), list(1))
+  expect_identical_(list2(!!!1i), list(1i))
+  expect_identical_(list2(!!!"foo"), list("foo"))
+  expect_identical_(list2(!!!bytes(0)), list(bytes(0)))
+})
+
+test_that("!!! calls as.list()", {
+  as_quos_list <- function(x, env = empty_env()) {
+    new_quosures(map(x, new_quosure, env = env))
+  }
+  exp <- as.list(mtcars)
+  expect_identical_(exprs(!!!mtcars), exp)
+  expect_identical_(quos(!!!mtcars), as_quos_list(exp))
+  expect_identical_(expr(foo(!!!mtcars)), do.call(call, c(list("foo"), exp)))
+  expect_identical_(list2(!!!mtcars), as.list(mtcars))
+
+  fct <- factor(c("a", "b"))
+  exp <- set_names(as.list(fct), c("", ""))
+  expect_identical_(exprs(!!!fct), exp)
+  expect_identical_(quos(!!!fct), as_quos_list(exp))
+  expect_identical_(expr(foo(!!!fct)), do.call(call, c(list("foo"), exp)))
+  expect_identical_(list2(!!!fct), as.list(fct))
+})
+
+test_that("!!! calls methods::as()", {
+  as_quos_list <- function(x, env = empty_env()) {
+    new_quosures(map(x, new_quosure, env = env))
+  }
+
+  .Person <- setClass("Person", slots = c(name = "character", species = "character"))
+  fievel <- .Person(name = "Fievel", species = "mouse")
+  methods::setAs("Person", "list", function(from, to) list(name = from@name, species = from@species))
+
+  exp <- list(name = "Fievel", species = "mouse")
+  expect_identical_(exprs(!!!fievel), exp)
+  expect_identical_(quos(!!!fievel), as_quos_list(exp))
+  expect_identical_(expr(foo(!!!fievel)), quote(foo(name = "Fievel", species = "mouse")))
+  expect_identical_(list2(!!!fievel), exp)
 })
 
 
@@ -316,7 +386,7 @@ test_that("double and triple ! are treated as syntactic shortcuts", {
 
 test_that("`!!` works in prefixed calls", {
   var <- quo(cyl)
-  expect_identical(expr_interp(~mtcars$`!!`(quo_expr(var))), ~mtcars$cyl)
+  expect_identical(expr_interp(~mtcars$`!!`(quo_squash(var))), ~mtcars$cyl)
   expect_identical(expr_interp(~foo$`!!`(quote(bar))), ~foo$bar)
   expect_identical(expr_interp(~base::`!!`(quote(list))()), ~base::list())
 })
@@ -348,6 +418,33 @@ test_that("can use prefix form of `!!` with qualifying operators", {
   expect_identical(expr(rlang:::`!!`(quote(bar))), quote(rlang:::bar))
 })
 
+test_that("can unquote within for loop (#417)", {
+  # Checks for an issue caused by wrong refcount of unquoted objects
+
+  x <- new_list(3)
+
+  for (i in 1:3) {
+    x[[i]] <- expr(!!i)
+  }
+  expect_identical(x, as.list(1:3))
+
+  for (i in 1:3) {
+    x[[i]] <- quo(!!i)
+  }
+  expect_identical(x, map(1:3, new_quosure, env = empty_env()))
+
+  for (i in 1:3) {
+    x[[i]] <- quo(foo(!!i))
+  }
+  exp <- list(quo(foo(1L)), quo(foo(2L)), quo(foo(3L)))
+  expect_identical(x, exp)
+
+  for (i in 1:3) {
+    x[[i]] <- quo(foo(!!!i))
+  }
+  expect_identical(x, exp)
+})
+
 
 # quosures -----------------------------------------------------------
 
@@ -373,10 +470,6 @@ test_that("can unquote-splice symbols", {
 
 test_that("can unquote symbols", {
   expect_error_(dots_values(!! quote(.)), "`!!` in a non-quoting function")
-
-  with_non_verbose_retirement(
-    expect_error_(dots_values(rlang::UQ(quote(.))), "`!!` in a non-quoting function")
-  )
 })
 
 
@@ -404,24 +497,82 @@ test_that("`:=` chaining is detected at dots capture", {
 test_that("Unquote operators fail when called outside quasiquoted arguments", {
   expect_qq_error <- function(object) expect_error(object, regexp = "within a quasiquoted argument")
   expect_qq_error(UQ())
-  expect_warning_(expect_qq_error(UQE()), "deprecated")
   expect_qq_error(UQS())
   expect_qq_error(`!!`())
   expect_qq_error(`!!!`())
   expect_qq_error(a := b)
 })
 
+test_that("`.data[[` unquotes", {
+  foo <- "bar"
+  expect_identical_(expr(.data[[foo]]), quote(.data[["bar"]]))
+  expect_identical_(expr(deep(.data[[foo]])), quote(deep(.data[["bar"]])))
+  expect_identical_(exprs(.data[[foo]]), named_list(quote(.data[["bar"]])))
+})
+
+test_that("it is still possible to unquote manually within `.data[[`", {
+  scoped_silent_retirement()
+  foo <- "baz"
+  expect_identical(expr(.data[[!!toupper(foo)]]), quote(.data[["BAZ"]]))
+})
+
+test_that(".data[[ argument is not masked", {
+  cyl <- "carb"
+  expect_identical_(eval_tidy(expr(.data[[cyl]]), mtcars), mtcars$carb)
+})
+
+test_that(".data[[ on the LHS of := fails", {
+  expect_error(exprs(.data[["foo"]] := foo), "Can't use the `.data` pronoun on the LHS")
+})
+
+test_that("it is still possible to use .data[[ in list2()", {
+  .data <- mtcars
+  expect_identical_(list2(.data$cyl), list(mtcars$cyl))
+})
+
 
 # Lifecycle ----------------------------------------------------------
 
-test_that("namespaced unquoting is soft-deprecated", {
-  with_non_verbose_retirement({
-    expect_no_warning_(exprs(rlang::UQS(1:2)))
-    expect_no_warning_(quo(list(rlang::UQ(1:2))))
-  })
+test_that("unquoting with rlang namespace is deprecated", {
+  expect_warning_(exprs(rlang::UQS(1:2)), regexp = "deprecated as of rlang 0.3.0")
+  expect_warning_(quo(list(rlang::UQ(1:2))), regexp = "deprecated as of rlang 0.3.0")
 
-  with_verbose_retirement({
-    expect_warning_(exprs(rlang::UQS(1:2)), "`UQS()` with a namespace is soft-deprecated", fixed = TRUE)
-    expect_warning_(quo(list(rlang::UQ(1:2))), "`UQ()` with a namespace is soft-deprecated", fixed = TRUE)
-  })
+  # Old tests
+
+  expect_identical(quo(rlang::UQ(toupper("a"))), new_quosure("A", empty_env()))
+  expect_identical(quo(list(rlang::UQS(list(a = 1, b = 2)))), quo(list(a = 1, b = 2)))
+
+  quo <- quo(rlang::UQ(NULL))
+  expect_equal(quo, ~NULL)
+
+  quo <- tryCatch(quo(rlang::UQ()), error = identity)
+  expect_is(quo, "error")
+  expect_match(quo$message, "must be called with an argument")
+
+  expect_error_(dots_values(rlang::UQ(quote(.))), "`!!` in a non-quoting function")
+})
+
+test_that("UQE() is defunct", {
+  expect_error_(expr(foo$UQE(NULL)), "defunct")
+  expect_error_(UQE(), "defunct")
+})
+
+test_that("splicing language objects still works", {
+  scoped_silent_retirement()
+
+  expect_identical_(exprs(!!!~foo), named_list(~foo))
+  expect_identical_(exprs(!!!quote(foo(bar))), named_list(quote(foo(bar))))
+
+  expect_identical_(quos(!!!~foo), quos_list(quo(!!~foo)))
+  expect_identical_(quos(!!!quote(foo(bar))), quos_list(quo(foo(bar))))
+
+  expect_identical_(expr(foo(!!!~foo)), expr(foo(!!~foo)))
+  expect_identical_(expr(foo(!!!quote(foo(bar)))), expr(foo(foo(bar))))
+
+  expect_identical_(list2(!!!~foo), list(~foo))
+  expect_identical_(list2(!!!quote(foo(bar))), list(quote(foo(bar))))
+})
+
+test_that("can unquote string in function position", {
+  expect_identical(expr((!!"foo")()), quote("foo"()))
 })
